/* ++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
/* +++               LittleShell.cpp                  +++ */
/* ++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
/* +++         Copyright © 2007 Ove Peter.            +++ */
/* ++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

#include <support/ClassInfo.h>
#include <storage/FilePanel.h>
#include <interface/TextView.h>
#include <storage/FindDirectory.h>
#include <iostream>

#include "ShellWindow.h"
#include "LittleShell.h"
#include "../olib/Colors.h"


int main( void )
{
    LittleShellApp *theApp;

    theApp = new LittleShellApp();
    theApp->Run();
	
    delete(theApp);
    return(0);
}


LittleShellApp::LittleShellApp( void ) : BApplication(APP_SIG)
{
	/* --- Set some objects to default values --- */
	isPanel = false;
	winNumber = 0;
	bubHelp = NULL;

	/* --- Initialize homeDir datamember --- */
	BPath	homepath;
	if(find_directory(B_USER_DIRECTORY, &homepath) == B_OK) {
		homeDir = homepath.Path();
	} else {
		homeDir = "/boot/home";
		cerr << "Error detecting user homedir path, setting to /boot/home ." << endl;
	}
	
	/* --- Check if settings file is present. --- */
	bool settings;
	if(prefs.FindBool("settings", &settings) != B_OK) {
		ThrowAlert("Welcome to "APP_NAME "!\n\n"
			"It seems to be the first time you're using this app.\n\n"
			"Please take a look at the preferences and read "
			"the manual, as not all features are obvious at first."
			"\nHave fun!\n"
			, "Don't show this again", B_IDEA_ALERT);
	} else {	// If present, check for version string.
		BString savedVersion;
		prefs.FindString("version", &savedVersion);
		if(savedVersion != APP_VERSION) {
			ThrowAlert("A settings file from a different version was found, using that"
				" (but resetting window dimensions).\n\n"
				"*If* anything should look or behave stangely, please quit the app "
				"and delete the old settings located in /boot/home/config/settings."
				,"OK", B_WARNING_ALERT);
				prefs.RemoveName("winsize");	// Reset win dimensions.
		}
	}
	
	/* --- Open first window --- */
	NewShellWindow();
}


LittleShellApp::~LittleShellApp( void )
{
	/* --- Save settings (present and app version) --- */
	prefs.Load();	// Reload prefs, as they could have changed meanwhile!
	bool settings = true;
	prefs.SetBool("settings", settings);
	prefs.SetString("version", APP_VERSION);
	prefs.Save();
	/* --- Delete FilePanel --- */
	if(isPanel)
		delete fOpenPanel;
}

/*-----------------------------------------------------------------------*/

void LittleShellApp::AboutRequested( void )
{
	int32			areply;
	BFont			alFont;
	BAlert			*alert;
	BTextView		*alText;
	ShellWindow		*activeWin = ActiveShellWin();
	BString			abouttext;

	abouttext << APP_NAME << " " << APP_VERSION << "\n\nBuilt on " << __DATE__ << " for BeOS ";
	abouttext << (int)(B_BEOS_VERSION >> 8) << "." << (int)((B_BEOS_VERSION >> 4) & 0xf) << "." << (int)(B_BEOS_VERSION & 0xf);
	#ifdef __INTEL__
		abouttext << " x86";
	#elif defined __POWERPC__
		abouttext << " PPC";
	#endif
	abouttext << "\n\nCopyright © 2007 Ove Peter\n\nDistributed under the MIT license.\nUse at your own risk. No Warranty!\n";

	alert = new BAlert(APP_NAME, abouttext.String(), "Web", "Email", "OK", B_WIDTH_AS_USUAL, B_INFO_ALERT);
	alert->SetShortcut(2, B_ESCAPE);
	
	// Make some stylings with the text view child of alert
	alText = alert->TextView();
	alText->SetStylable(true);
			
	alFont.SetSize(18);
	alFont.SetFace(B_BOLD_FACE);
	alText->SetFontAndColor(0, 23, &alFont);
	alFont.SetSize(16);
	alText->SetFontAndColor(12, 28, &alFont, B_FONT_ALL, &blueCol);
	
	activeWin->Lock();
	areply = alert->Go();
	
	if(areply == 0)
		activeWin->LaunchURLs(HOMEPAGE, "http");
	else if(areply == 1)
		activeWin->LaunchURLs("mailto:ovepeter@gmx.de", "email");
	
	activeWin->Unlock();
}


void LittleShellApp::MessageReceived(BMessage *message)
{
	#ifdef DEBUG
		message->PrintToStream();
	#endif

	switch(message->what)
	{
		case OPEN_FILES:
			if(!isPanel) {	// first OpenPanel request, get workDir
				isPanel = true;
				entry_ref wdref;
				message->FindRef("workdir", &wdref);	// No check needed, file panel will default to home if not valid
				fOpenPanel = new BFilePanel(B_OPEN_PANEL, NULL, &wdref, B_FILE_NODE | B_DIRECTORY_NODE, true, new BMessage(FILE_PANEL_MSG));
			}
			fOpenPanel->Show();
			break;
		
		case FILE_PANEL_MSG:
		{	// Message from OpenPanel
			ShellWindow		*activeWin = ActiveShellWin();
			activeWin->Lock();
			activeWin->AddFileRefs(message);
			activeWin->Unlock();
		}
			break;
			
		case B_SILENT_RELAUNCH:
			NewShellWindow();
			break;
				
		case NEW_SHELL_WINDOW:
			NewShellWindow();
			break;
		
		case UPDATE_HIST_MENU:
		{
			BString	newCommand;
			message->FindString("newcommand", &newCommand);
						
			for(int32 i=0; i<CountWindows(); i++) {
				if(is_kind_of(WindowAt(i), ShellWindow)) {
					// WindowAt(i)->Lock();
					((ShellWindow *)WindowAt(i))->UpdateHistory(newCommand.String());
					// WindowAt(i)->Unlock();
				}
			}
		}
			break;
		
		case UPDATE_MAX_HIST:
		{
			int32	maxHistItems;
			message->FindInt32( "maxhistory", 0, &maxHistItems);
			
			for(int32 i=0; i<CountWindows(); i++)
				if(is_kind_of(WindowAt(i), ShellWindow))
					((ShellWindow *)WindowAt(i))->UpdateMaxHist(maxHistItems);
		}
			break;

		case UPDATE_WIN_LOOK:
		{
			BString	winLook;
			message->FindString("windowlook", &winLook);

			for(int32 i=0; i<CountWindows(); i++)
				if(is_kind_of(WindowAt(i), ShellWindow))
					((ShellWindow *)WindowAt(i))->SetTheWinType(winLook.String());
		}
			break;
		
		case UPDATE_ADD_AMPER:
		{
			bool addAmper;
			message->FindBool("upamper", &addAmper);
			
			for(int32 i=0; i<CountWindows(); i++)
				if(is_kind_of(WindowAt(i), ShellWindow))
					((ShellWindow *)WindowAt(i))->UpdateAddAmper(addAmper);			
		}
			break;

		case UPDATE_STDOUT:
		{
			bool showStdout;
			message->FindBool("stdout", &showStdout);
			
			for(int32 i=0; i<CountWindows(); i++)
				if(is_kind_of(WindowAt(i), ShellWindow))
					((ShellWindow *)WindowAt(i))->UpdateShowOut(showStdout);			
		}
			break;

		case UPDATE_OUTTITLE:
		{
			bool outTitle;
			message->FindBool("stdouttitle", &outTitle);
			
			for(int32 i=0; i<CountWindows(); i++)
				if(is_kind_of(WindowAt(i), ShellWindow))
					((ShellWindow *)WindowAt(i))->UpdateOutTitle(outTitle);
		}
		
		case UPDATE_STAY_TOP:
		{
			bool stayTop;
			message->FindBool("stayontop", &stayTop);
			
			for(int32 i=0; i<CountWindows(); i++)
				if(is_kind_of(WindowAt(i), ShellWindow))
					((ShellWindow *)WindowAt(i))->UpdateStayOnTop(stayTop);
		}
			break;

		case UPDATE_AUTOCLOSE:
		{
			bool autoClose;
			message->FindBool("autoclose", &autoClose);
			
			for(int32 i=0; i<CountWindows(); i++)
				if(is_kind_of(WindowAt(i), ShellWindow))
					((ShellWindow *)WindowAt(i))->UpdateAutoClose(autoClose);
		}
			break;
		
		case UPDATE_SHOW_WDIR:
		{
			bool showWorkDir;
			message->FindBool("wdirtitle", &showWorkDir);
			
			for(int32 i=0; i<CountWindows(); i++)
				if(is_kind_of(WindowAt(i), ShellWindow))
					((ShellWindow *)WindowAt(i))->UpdateShowWDir(showWorkDir);
		}
			break;
		
		case UPDATE_SAVE_WDIR:
		{
			bool saveWorkDir;
			message->FindBool("saveworkdir", &saveWorkDir);
			
			for(int32 i=0; i<CountWindows(); i++)
				if(is_kind_of(WindowAt(i), ShellWindow))
					((ShellWindow *)WindowAt(i))->UpdateSaveWDir(saveWorkDir);
		}
			break;

		default:
			BApplication::MessageReceived(message);
			break;
	}
}


void LittleShellApp::RefsReceived(BMessage *message)
{	
	// If only one window is open (started by the add-on likely) and it is in current workspace, 
	// post the message to it. When more than one window, spawn a new one.

	int32 totalWins = CountWindows();
	int32 last = 0;
	int32 totalShell = 0;
	
	/* --- Find last window in current workspace (searching an active one won't help in some cases) --- */
	for(int32 i=0; i<totalWins; i++) {
		if( is_kind_of(WindowAt(i), ShellWindow) && (WindowAt(i)->Workspaces() & (uint32)(1 << current_workspace())) )
		{
			last = i;
			totalShell++;
		}
	}
			
	if(totalShell != 1)
	{			/* --- Create new Window --- */
		ShellWindow	*newWindow = NewShellWindow();
		
		newWindow->Lock();
		newWindow->AddFileRefs(message);
		newWindow->Unlock();
		newWindow->Activate();
	} else {	/* --- Use last window if the only one --- */
		WindowAt(last)->Lock();
		((ShellWindow *)WindowAt(last))->AddFileRefs(message);
		WindowAt(last)->Unlock();
		WindowAt(last)->Activate();
	}
}

/* ======================================================================================= */

ShellWindow* LittleShellApp::NewShellWindow( void )
{
	ShellWindow		*fAWindow;
	BRect			aRect;
	int32			totalWins = CountWindows();
	
	winNumber++;
	
	if(totalWins == 0)
	{			/* --- First window, get saved or default size --- */
		if(prefs.FindRect("winsize", &aRect) != B_OK)
			aRect.Set(WIN_LEFT, WIN_TOP, (WIN_WIDTH + WIN_LEFT), (WIN_HEIGHT + WIN_TOP));
	} else {	/* --- Get Frame of last Shell Window (isn't necessarily active) and open new window near to it --- */
		int32	last = 0;
		for(int32 i=0; i<totalWins; i++) {
			if( is_kind_of(WindowAt(i), ShellWindow) && (WindowAt(i)->Workspaces() & (uint32)(1 << current_workspace())) ) {
				last = i;
			}
		}
		aRect = WindowAt(last)->Frame();
		aRect.OffsetBy(15.0, 15.0);
	}

	fAWindow = new ShellWindow(aRect, winNumber, homeDir);
	return fAWindow;
}


ShellWindow* LittleShellApp::ActiveShellWin( void )
{
	int32	totalWins = CountWindows();
	int32	currWin = 0; 	/* assigning a value is safer, a window index will be
							supplied nevertheless in case IsActive() below fails. */
	for(int32 i=0; i<totalWins; i++) {
		if( is_kind_of(WindowAt(i), ShellWindow) ) {
			if( WindowAt(i)->IsActive() == true ) {
				currWin = i;
				break;
			}
		}
	}

	return((ShellWindow *)WindowAt(currWin));
}


void LittleShellApp::SetBubbleHelp(BView *view, char *text)
{
	/* Create the BubbleHelper first if not existant and if text is NOT NULL!
	(NULL means erase text of running BubbleHelper, but it doesn't run
	*/
	if( (bubHelp == NULL) && (text != NULL) ){
		bubHelp = new BubbleHelper();
	}
	
	// This will never be executed when window started with BubbleHelp disabled
	if(bubHelp != NULL) {
		bubHelp->SetHelp(view, text);
	}	
}


void LittleShellApp::ThrowAlert(const char *text, const char *buttonlabel, alert_type type)
{
	BAlert		*alert;
	alert = new BAlert("alert", text, buttonlabel, NULL, NULL, B_WIDTH_FROM_LABEL, type);
	alert->Go();
}
