/* ++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
/* +++               ShellWindow.cpp                  +++ */
/* ++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
/* +++         Copyright © 2007 Ove Peter.            +++ */
/* ++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

#include <app/Application.h>
#include <app/Roster.h>
#include <support/ClassInfo.h>
#include <storage/FindDirectory.h>
#include <interface/Screen.h>
#include <interface/TextControl.h>
#include <interface/MenuItem.h>
#include <interface/MenuField.h>
#include <interface/Button.h>
#include <stdlib.h>		// for system()
#include <unistd.h>		// for getcwd(), chdir()
#include <limits.h>		// for PATH_MAX
#include <iostream>


#include "ShellWindow.h"
#include "LittleShell.h"
#include "PrefWindow.h"
#include "../olib/OMenu.h"
#include "../olib/OMenuBar.h"
#include "../olib/Colors.h"

/*----------------- Constructor & Deconstructor -----------------------*/

ShellWindow::ShellWindow(BRect frame, uint32 windowNumber, BString homedir)
    : BWindow(frame, APP_NAME, DEF_WIN_TYPE, B_NOT_V_RESIZABLE | B_ASYNCHRONOUS_CONTROLS)
{	
	homeDir = homedir;
	winNum = windowNumber;
	hist_busy = false;

	/* ------------------ Window size and position ------------------- */
	
	// Set Size Limits
	float minWidth, maxWidth, minHeight, maxHeight;
    GetSizeLimits(&minWidth, &maxWidth, &minHeight, &maxHeight);
    SetSizeLimits(200.0, maxWidth, WIN_HEIGHT, WIN_HEIGHT);
	
	// Make sure the window is not sized offscreen	
	BScreen theScreen;
	BRect	wholeScreen = theScreen.Frame();
	BRect	windowSize;
	float	winWidth, winHeight;

	windowSize = frame;
	if((windowSize.left > wholeScreen.Width()) || (windowSize.top > wholeScreen.Height()))
	{
		MoveTo(WIN_LEFT, WIN_TOP);
		winWidth = WIN_WIDTH;
		winHeight = WIN_HEIGHT;
	} else {
		MoveTo(windowSize.left, windowSize.top);
		winWidth = windowSize.Width();
		winHeight = windowSize.Height();
	}
	ResizeTo(winWidth, winHeight);

	/*------------------------- Load settings -----------------------------*/

	// GetLastCommand
	if(prefs.FindString("lastcommand", &lastCommand) != B_OK)
		lastCommand = HOMEPAGE;
	
	// Restrict history to maxHistItems entries
	if(prefs.FindInt32("maxhistory", &maxHistItems) != B_OK)
		maxHistItems = MAX_HIST_ITEMS;
		
	// Close window after command is done or stay open 
	if(prefs.FindBool("autoclose", &autoClose) != B_OK)
		autoClose = false;

	// Wether to open a window for stdout of shell commands
	if(prefs.FindBool("showstdout", &showStdout) != B_OK)
		showStdout = true;

	// Wether to show the command as stdout window title
	if(prefs.FindBool("stdouttitle", &showOutTitle) != B_OK)
		showOutTitle = true;
	
	// Wether to append an ampersand to shell commands
	if(prefs.FindBool("addamper", &addAmper) != B_OK)
		addAmper = false;
	
	// Wether to stay On Top (will be set by SetTheWinType() also)
	if(prefs.FindBool("stayontop", &stayOnTop) != B_OK)
		stayOnTop = false;
	
	// Wether to show Bubble Help
	if(prefs.FindBool("bubblehelp", &showBubble) != B_OK)
		showBubble = true;
	
	// Wether to show the Work Directory in Title
	if(prefs.FindBool("wdirtitle", &showWorkDir) != B_OK)
		showWorkDir = true;
	
	// Wether to save the last WorkDir
	if(prefs.FindBool("saveworkdir", &saveWorkDir) != B_OK)
		saveWorkDir = true;
	
	// Get/Set workdir BString
	if(saveWorkDir)
	{
		if(prefs.FindString("lastworkdir", &workDir) != B_OK)
			workDir = homeDir;
	} else {	// Get current app PWD (unistd.h) when option to save workdir is off
		char *buff = NULL;
		char *theDirBuf;
		if( (theDirBuf = getcwd(buff, PATH_MAX)) != NULL )
			workDir << theDirBuf;
	}
	// Check path (also sets it if valid, and updates title, etc...)
	if(WorkDirCheckSet(workDir, true) != B_OK) {
		cerr << "Last used workdir invalid, resetting to homedir." << endl;
		WorkDirCheckSet(homeDir, true);
	}
	
	// Get Window look
	if(prefs.FindString("windowlook", &winType) != B_OK)
		winType = "Standard";

	SetTheWinType(winType.String());	// Set the Window Type and Feel!

	/* --------------- Add the BMenuBar, etc. to window ---------------- */
	
	BRect		menuBarRect;
	menuBarRect = Bounds();
	menuBarRect.bottom = 18.0;
	
	// OMenuBar defaults to not automatically resize, otherwise it's all fucked up with very small/big fonts.
	menuBar = new OMenuBar(menuBarRect, "menu_bar", B_FOLLOW_LEFT_RIGHT | B_FOLLOW_TOP);
	
	/* File menu */
	OMenu		*fileMenu = new OMenu("File", menuBar);
	BMenuItem	*fileMI;
	
	fileMenu->AddItem(openFileMI = new BMenuItem("Open Files..", new BMessage(OPEN_FILES), 'O', B_COMMAND_KEY));
	fileMenu->AddItem(fileMI = new BMenuItem("Tracker Find..", new BMessage(FIND_PANEL), 'F', B_COMMAND_KEY));
	fileMI->SetTarget(TRACKER_SIG);		// Message to Tracker (open FindPanel)
	fileMenu->AddSeparatorItem();		// WAS: SetTarget(BMessenger(TRACKER_SIG))
	fileMenu->AddItem(fileMI = new BMenuItem("About LittleShell", new BMessage(B_ABOUT_REQUESTED), 'A', B_COMMAND_KEY | B_SHIFT_KEY));
	fileMI->SetTarget(be_app);			// Message to LS app
	fileMenu->AddSeparatorItem();
	fileMenu->AddItem(fileMI = new BMenuItem("Quit app", new BMessage(QUIT_APP), 'Q', B_COMMAND_KEY));
	
	/* Standard Edit menu */
	OMenu		*editMenu = new OMenu("Edit", menuBar);
	BMenuItem	*textEditMI;
	
	editMenu->AddItem(textEditMI = new BMenuItem("Undo Text", new BMessage(B_UNDO), 'Z', B_COMMAND_KEY));
	textEditMI->SetTarget(NULL, this);
	editMenu->AddSeparatorItem();
	editMenu->AddItem(textEditMI = new BMenuItem("Cut", new BMessage(B_CUT), 'X', B_COMMAND_KEY));
	textEditMI->SetTarget(NULL, this);
	editMenu->AddItem(textEditMI = new BMenuItem("Copy", new BMessage(B_COPY), 'C', B_COMMAND_KEY));
	textEditMI->SetTarget(NULL, this);
	editMenu->AddItem(textEditMI = new BMenuItem("Paste", new BMessage(B_PASTE), 'V', B_COMMAND_KEY));
	textEditMI->SetTarget(NULL, this);
	editMenu->AddSeparatorItem();
	editMenu->AddItem(textEditMI = new BMenuItem("Select All", new BMessage(B_SELECT_ALL), 'A', B_COMMAND_KEY));
	textEditMI->SetTarget(NULL, this);

	/* Options menu */
	OMenu		*optionMenu = new OMenu("Options", menuBar);
		
	optionMenu->AddItem(autoCloseMI = new BMenuItem("Auto Close", new BMessage(AUTO_CLOSE), 'W', B_COMMAND_KEY | B_SHIFT_KEY));
	autoCloseMI->SetMarked(autoClose);
	optionMenu->AddItem(stayOnTopMI = new BMenuItem("Stay On Top", new BMessage(STAY_ON_TOP), 'T', B_COMMAND_KEY));
	stayOnTopMI->SetMarked(stayOnTop);
	optionMenu->AddItem(showOutMI = new BMenuItem("Show stdout", new BMessage(SHOW_STDOUT), 'S', B_COMMAND_KEY));
	showOutMI->SetMarked(showStdout);
	// Next item will be updated with correct status later by SetBubbleHelp()
	optionMenu->AddItem(bubbleMI = new BMenuItem("Bubble Help", new BMessage(BUBBLE), 'B', B_COMMAND_KEY | B_SHIFT_KEY));
	optionMenu->AddSeparatorItem();
	BMenuItem	*prefsMI;
	optionMenu->AddItem(prefsMI = new BMenuItem("Preferences...", new BMessage(OPEN_PREF_WIN), 'P', B_COMMAND_KEY));
	
	/* Window menu */
	OMenu		*winMenu = new OMenu("Window", menuBar);
	BMenuItem	*winMI;
		
	winMenu->AddItem(winMI = new BMenuItem("New win", new BMessage(NEW_SHELL_WINDOW), 'N', B_COMMAND_KEY));
	winMenu->AddItem(winMI = new BMenuItem("Zoom", new BMessage(ZOOM_NOW), 'Y', B_COMMAND_KEY));
	winMenu->AddItem(winMI = new BMenuItem("Minimize", new BMessage(MINI_NOW), 'M', B_COMMAND_KEY));
	winMenu->AddSeparatorItem();
	winMenu->AddItem(winMI = new BMenuItem("Close", new BMessage(B_QUIT_REQUESTED), 'W', B_COMMAND_KEY));	

	// Add all BMenu's to BMenuBar...
	menuBar->AddItem(fileMenu);
	menuBar->AddItem(editMenu);
	menuBar->AddItem(optionMenu);
	menuBar->AddItem(winMenu);
	AddChild(menuBar);
	
	/*----------- Construct the top view with all controls ---------------*/
	
	BRect	backRect = Bounds();
	backRect.top = 	(menuBarRect.bottom + 1.0);
	backRect.bottom = (winHeight - 14.5);
	// Top View
    fAView = new TopView(backRect, "LittleShellView");

	// As the Button will be default, we need to subtract ~4 points for Be's "default border"
	BRect	butRect((backRect.Width() - 31), 4, (backRect.Width() - 4), (backRect.Height() - 4));
	okButt = new BButton(butRect, "OK-B", "OK", new BMessage(GO_BUTTON), B_FOLLOW_RIGHT, B_WILL_DRAW | B_NAVIGABLE );
	okButt->SetFont(be_fixed_font);
	okButt->SetFontSize(12.0);
	okButt->MakeDefault(true);
    fAView->AddChild(okButt);
    
	// Command Input Bar
	BRect barRect(28, 5, (backRect.Width() - 40), 28);
	commandBar = new OTextControl(barRect, "CommandInput", NULL, lastCommand.String(), NULL);
	fAView->AddChild(commandBar);
	
	/* ---- Stick History pop-up into view ----- */
	BRect 		histRect(5, 5, 22 ,25);
	const char	*historyCommand;
	
	historyMenu = new BMenu(" ");

	if(prefs.FindString("history", &historyCommand) == B_OK) {		
		BMenuItem	*histMI;
		for(int32 i=0; prefs.FindString("history", i, &historyCommand) == B_OK; i++) {
			if(historyCommand) {
			if(historyCommand[0]) {
				// Add to autocomplete list
				commandBar->AddAutoText(new BString(historyCommand));
				// Add to Menu
				BMessage *histmsg = new BMessage(HIST_ITEM_SELECTED);
				histmsg->AddString("history", historyCommand);
				historyMenu->AddItem(histMI = new BMenuItem(historyCommand, histmsg));
				// TODO: color every 2nd item different
			}
			}
		}
		// Set Shortcut ALT-L for first MenuItem (lastcommand)
		historyMenu->ItemAt(0)->SetShortcut('L', B_COMMAND_KEY);
	}

	histMenuField = new BMenuField(histRect, "HistoryPopUp", NULL, historyMenu, true, B_FOLLOW_LEFT | B_FOLLOW_TOP, B_WILL_DRAW | B_NAVIGABLE | B_NAVIGABLE_JUMP);
	fAView->AddChild(histMenuField);
	historyMenu->SetFont(be_plain_font); // for this menu, SetFont() has to be called after AddChild().

	AddChild(fAView);

	/*----------------------- Add the status views ----------------------*/
	
	BRect	statusRect = Bounds();
	statusRect.bottom = winHeight;
	statusRect.top = (winHeight - 13.0);	
	
	statusView = new StatusView(statusRect, "statusView", yelCol);
	AddChild(statusView);
	statusView->NewText(" ...");

	/* ------- Make final adjustments  -------- */
	commandBar->SelectAll();
	commandBar->MakeFocus(true);
	
	/* ---- Fill the BubbleHelper with text after all other childs are attached  --- */
	SetBubbleHelp(showBubble);
	
	/* --- Make window visible --- */
	Show();
}


ShellWindow::~ShellWindow( void )
{

}


/*--------------------------------------------------------------------------*/


bool ShellWindow::QuitRequested( void )
{
	/* --- Save settings! --- */
	prefs.SetRect("winsize", Frame());
	prefs.SetBool("autoclose", autoClose);
	prefs.SetBool("stayontop", stayOnTop);
	prefs.SetBool("showstdout", showStdout);
	prefs.SetBool("stdouttitle", showOutTitle);
	prefs.SetBool("wdirtitle", showWorkDir);
	prefs.SetInt32("maxhistory", maxHistItems);
	prefs.SetBool("bubblehelp", showBubble);
	prefs.SetBool("saveworkdir", saveWorkDir);
	if(saveWorkDir) prefs.SetString("lastworkdir", workDir.String());
	prefs.Save();

	/* --- If this is the last ShellWindow, clean everything up --- */
	int32 totalWindows = 0;
	for(int32 i=0; i<be_app->CountWindows(); i++)
		if(is_kind_of(be_app->WindowAt(i), ShellWindow))
			totalWindows++;

	if(totalWindows == 1)
		be_app->PostMessage(B_QUIT_REQUESTED);

	return(true);
}


void ShellWindow::MessageReceived(BMessage *message)
{	
	#ifdef DEBUG
		message->PrintToStream();
	#endif

	switch(message->what)
	{
		case B_SIMPLE_DATA:
			// On drop-event, look for a ref in message
			AddFileRefs(message);
			break;

		case GO_BUTTON:
			// Here we execute finally:
			Execute(message);
		  	break;
		
		case OPEN_FILES:
		{	// Send an entry_ref containing workDir on first OpenPanel
			BMessage 	*msg = new BMessage(OPEN_FILES);
			entry_ref	wdref;
			BEntry		wdent(workDir.String(), true);

			wdent.GetRef(&wdref);
			msg->AddRef("workdir", &wdref);
			
			be_app->PostMessage(msg);
			delete msg;
			// Now change the Menu Item's msg Handler directly to be_app,
			// so the above isn't repeatedly done
			openFileMI->SetTarget(NULL, be_app);
		}
			break;
			
		case OPEN_PREF_WIN:
			OpenPrefWin();
			break;

		case HIST_ITEM_SELECTED:
			const char *newCommand;
			if(message->FindString("history", &newCommand) == B_OK)
			{
				commandBar->SetText(newCommand);
				commandBar->MakeFocus(true);
			}
			break;
		
		case NEW_SHELL_WINDOW:
			be_app->PostMessage(NEW_SHELL_WINDOW); 
			break;

		case ZOOM_NOW:
			Zoom();
			break;
			
		case MINI_NOW:
			// If stay on top is on, we need to turn it off first to minimize the window.
			if(Feel() == B_FLOATING_ALL_WINDOW_FEEL)
				UpdateStayOnTop(false);
			Minimize(true);
			break;

		case AUTO_CLOSE:
			// Set check state of menu item
			// bool autoClose will be checked after execution.
			if(autoClose == true) UpdateAutoClose(false);
			else UpdateAutoClose(true);
			// Save Settings (-> see comment!)
			prefs.SetBool("autoclose", autoClose);
			prefs.Save();	// Remove Save() as all options are integrated also as global pref ???!!
			break;			// Would result in new windows not getting the settings of current window, but
							// it would follow the concept of global prefs in pref window vs. individual window prefs.
		case STAY_ON_TOP:
			if(stayOnTop == true) UpdateStayOnTop(false);
			else UpdateStayOnTop(true);	// this function takes care of everything
			// Save prefs for new windows
			prefs.SetBool("stayontop", stayOnTop);
			prefs.Save();
			break;
			
		case SHOW_STDOUT:
			if(showStdout == true) UpdateShowOut(false);
			else UpdateShowOut(true);
			prefs.SetBool("showstdout", showStdout);
			prefs.Save();
			break;

		case BUBBLE:
			if(showBubble == true) SetBubbleHelp(false);
			else SetBubbleHelp(true);
			prefs.SetBool("bubblehelp", showBubble);
			prefs.Save();
			break;
			
		case STATUSTEXT:
		{	/* --- Got a message to display in statusbar --- */
			const char	*stext;
			if( message->FindString("stattext", &stext) == B_OK ) {
				// iserror is optional, indicates error condition. If not set, leave as is.
				rgb_color	col = statusView->CurrentColor();
				bool		iserror;
				if( message->FindBool("iserror", &iserror) == B_OK ) {
					if(iserror == true) col = redCol;
					else col = greenCol;
				}
				// append is optional, default is stattext will be replaced
				bool	append = false;
				message->FindBool("append", &append);

				DisplayStatus(stext, col, append);
			} else {	// No statustext in message is not good!
				DisplayStatus(" Warning: received invalid status message", redCol, true);
			}
		}
			break;

		case QUIT_APP:
			// Close everything.
			be_app->PostMessage(B_QUIT_REQUESTED);
			break;
		
		default:
			BWindow::MessageReceived(message);
	}
}

/* =========================================================================================== */

status_t ShellWindow::WorkDirCheckSet(BString dirToCheck, bool alsoSet)
{
	if(dirToCheck.CountChars() == 0) return B_ERROR;	// Don't continue if given string is empty.
	
	// no "/" at beginning == relative from current workDir, prepend workDir
	if(dirToCheck.FindFirst("/", 0) != 0)
	{
		dirToCheck.Prepend("/");
		dirToCheck.Prepend(workDir);
	}
	
	BEntry	dirEntry;
	dirEntry.SetTo(dirToCheck.String(), true);
	
	if(dirEntry.IsDirectory() == true) {	
		
		if(alsoSet == true) {
			workDir = dirToCheck;			// Update the global datamember
			UpdateShowWDir(showWorkDir);	// Update the Window Title
			// Add to Recent Folders
			entry_ref ref;
			if(dirEntry.GetRef(&ref) == B_OK) be_roster->AddToRecentFolders(&ref, NULL);
			/* chdir(workDir.String());		--->> Change app dir (unistd.h)
			NOTE: chdir blocks unmounting volumes. It will be cd'd and checked anyway on execution. */
		}
		
		return B_OK;
	}

	return B_ERROR;			// It isn't a directory.
}


void ShellWindow::AddFileRefs(BMessage *message)
{
	#ifdef DEBUG
		message->PrintToStream();
	#endif
	
	DisplayStatus("Processing Files...", yelCol);
	
	int32		i;
	BString		tempDir;
	BString		FileNames;
	BPath		path;
	entry_ref	ref;
	BString		refstatus;
	bool		filenameonly = (modifiers() & B_SHIFT_KEY);
	bool		appendrefs = (modifiers() & B_LEFT_CONTROL_KEY);
	bool		prependrefs = (modifiers() & B_COMMAND_KEY);
	
	if(message->FindRef("refs", &ref) == B_OK)	// Look for a file ref in the message
	{
   		for( i = 0; message->FindRef("refs", i, &ref) == B_OK; i++ )
	   	{
			// Initialize path argument with full pathname ["path.Path" retrieves it]
   			BEntry(&ref).GetPath(&path);
			if(i>0) FileNames.Append(" ");	// If more than one file, put a space in between
			BString	FileString	= filenameonly ? path.Leaf() : path.Path();
		
			// Make sure that filenames are always encapsulated in quotes!
			// NOTE: Could make sense to use BString::CharacterEscape() here instead...
			FileNames.Append("\"").Append(FileString).Append("\"");
		}

		tempDir = path.Path();				// Get a reference to the directory from the last file
		tempDir.RemoveLast(path.Leaf());	// getting dirpath from filepath requires to strip off the filename

		if(appendrefs) {			// Insert at the end of TextControl
			FileNames.Prepend(" ");
			commandBar->AppendText(FileNames);
		} else if(prependrefs) {	// Insert at start
			FileNames.Append(" ");
			commandBar->PrependText(FileNames);
		} else {					// change all the text
			commandBar->SetText( FileNames.String() );
		}
		
		if(WorkDirCheckSet(tempDir) == B_OK) refstatus = "File refs added, workdir changed.";
		else refstatus = "File refs added, but path is invalid as workdir.";
		
	} else if(message->FindRef("dirref", &ref) == B_OK)	{	// Search for a special dirref name content from the Tracker add-on!
		BEntry(&ref).GetPath(&path);
		tempDir = path.Path();
		if(WorkDirCheckSet(tempDir) == B_OK) refstatus = "Work directory changed.";
		else refstatus = "Path is invalid, cancelling.";
	} else {	// Invalid data !!
		DisplayStatus("Error: data is no file or dir ref.", redCol);
		return;
	}
	
	DisplayStatus(refstatus.String(), yelCol);
}


void ShellWindow::DisplayStatus(const char *statText, const rgb_color statCol, bool append)
{
	if(append == true) {
		statusView->AppendText(statText);
	} else {
		statusView->NewText(statText);
	}

	statusView->NewCol(statCol);
}


void ShellWindow::Execute(BMessage *message)
{	// TODO: why not start this function in a new thread?
	BString		commandString;
	bool		reverthome = false;

	DisplayStatus("Working...", yelCol);
	
	// Get text in CommandBar
	commandString = (const char *)commandBar->Text();
	
	// Only continue if BString isn't empty, or FindFirst() crashes
	if( commandString.CountChars() == 0 ) {
		DisplayStatus("Error: No command given.", redCol);
		return;
	}
	
	// Disable controls	(after the possible return above) and sync window.
	commandBar->SetEnabled(false);
	histMenuField->SetEnabled(false);
	okButt->SetEnabled(false);
	
	// Replace "~" with homedir first (will be eventually reverted on save)
	if(commandString.FindFirst("~", 0) != B_ERROR) {
		reverthome = true;	
		commandString.ReplaceAll("~", homeDir.String());
	}
	
	// Go through the Check/Execute mechanisms
	if(SearchCD(commandString) == B_OK);					// -> got a single "cd " command
	else if(SearchURLs(commandString) == B_OK);				// -> got an URL
	else if(SearchTrackerFolder(commandString) == B_OK);	// -> got a directory
	else if(SearchFile(commandString) == B_OK);				// -> got a filepath
	else ExecuteShellCommand(commandString);				// -> got a shell command

	// Revert conversion of "~" to homedir to avoid confusion for the user
	if(reverthome == true)
		commandString.ReplaceAll(homeDir.String(), "~");
		
	/*-------------- Update history menus of all windows --------------*/
	
	// Set history-is-busy boolean flag
	hist_busy = true;

	// Tell the app to add the new word into the History menu of
	// each currently open shell window (if more than one). 
	// This keeps the windows in sync.
	if(be_app->CountWindows() > 1) {
		BMessage *updateHistMsg = new BMessage(UPDATE_HIST_MENU);
		updateHistMsg->AddString("newcommand", commandString);
		if(be_app->PostMessage(updateHistMsg) == B_OK)
			delete updateHistMsg;
	} else {
		UpdateHistory(commandString.String());
	}
	
	// Wait until history menu is updated by be_app.
	while(hist_busy) {
		snooze(10000);
	}
	
	// Quit or stay open, dependant on user choice
	if(autoClose) {
		PostMessage(B_QUIT_REQUESTED);
	} else { // ENTER Message (for commbar) gets stuck in default button here, so select all manually
		commandBar->SelectAll();		
	}

	// Re-enable controls!
	okButt->SetEnabled(true);
	commandBar->SetEnabled(true);
	histMenuField->SetEnabled(true);
}


status_t ShellWindow::SearchCD(BString commandString)
{
	if(commandString == "cd") {
	/* -- Only cd == switch to homedir -- */
		if(WorkDirCheckSet(homeDir) == B_OK) DisplayStatus("Work directory set to user homedir.", greenCol);
		return B_OK;				// return (cd command found)
	} else if(commandString.FindFirst("cd ", 0) == 0) {
	/* -- Trigger in first char -- */
		commandString.RemoveFirst("cd ");
		if(commandString.FindFirst("\"") != B_ERROR) commandString.RemoveAll("\""); // Strip off all quotes, if any		
		
		// Check if more than only a cd command. If so, change workdir but pass on the command.
		if(commandString.FindFirst(";", 0) != B_ERROR) {
			int32 offset = commandString.FindFirst(";", 0);
			BString commStr;
			commandString.CopyInto(commStr, 0, offset);
			if(WorkDirCheckSet(commStr) == B_OK) DisplayStatus("Work Dir changed.", greenCol);
			return B_ERROR;			// return (No cd command)
		}
		
		// Only a cd command, check given path and set workdir.
		if(commandString.FindFirst("..") != B_ERROR) {		// TODO: make "cd .." work!
			DisplayStatus("Sorry, paths with \"..\" in between are not allowed for workdir switching, only in shell commands.", redCol);
		} else if(WorkDirCheckSet(commandString) == B_OK) {	// Is a valid dir path (workDir datamember was set by WDCS())
			DisplayStatus("Work Dir changed.", greenCol);
		} else {
			DisplayStatus("cd path invalid (does not exist).", redCol);
		}
		return B_OK;
	}
	
	return B_ERROR;
}


status_t ShellWindow::SearchURLs(BString urlString)
{	// Search for Strings that are at the *first* character
	// (URL strings in the middle of the commandline should be omitted)
	if(urlString.IFindFirst("url://", 0) == 0)
	{	// Trigger for general URL found
		urlString.RemoveFirst("url://");
		LaunchURLs(urlString, "http");
	// For URLs, it's OK to search case-insensitive:
	} else if(urlString.IFindFirst("mailto:", 0) == 0) {
		LaunchURLs(urlString, "email");
	} else if(urlString.IFindFirst("http://", 0) == 0) {
		LaunchURLs(urlString, "http");
	} else if(urlString.IFindFirst("https://", 0) == 0) {
		LaunchURLs(urlString, "https");
	} else if(urlString.IFindFirst("www.", 0) == 0) {
		urlString.Prepend("http://");
		LaunchURLs(urlString, "http");
	} else if(urlString.IFindFirst("ftp://", 0) == 0) {
		LaunchURLs(urlString, "ftp");
	} else if(urlString.IFindFirst("ftp.", 0) == 0) {
		urlString.Prepend("ftp://");
		LaunchURLs(urlString, "ftp");
	} else if(urlString.IFindFirst("beshare://", 0) == 0) {
		urlString.RemoveAll("beshare://");
		LaunchURLs(urlString, "beshare");
	} else if(urlString.IFindFirst("rtsp://", 0) == 0) {
		LaunchURLs(urlString, "rtsp");
	} else if(urlString.IFindFirst("news://", 0) == 0) {
		LaunchURLs(urlString, "news");
	} else if(urlString.IFindFirst("telnet://", 0) == 0) {
		LaunchURLs(urlString, "telnet");
	} else if(urlString.IFindFirst("audio://", 0) == 0) {
		urlString.RemoveFirst("audio://");	// audiostreams start with http
		if(urlString.IFindFirst("http://", 0) != 0) urlString.Prepend("http://");
		LaunchURLs(urlString, "audiostream");
	} else {	// return NO URL FOUND
		return B_ERROR;
	}
		
	return B_OK;	// return URL found
}


status_t ShellWindow::LaunchURLs(BString urlString, BString type)
{
	status_t	roster_stat = B_ERROR;
	char 		*urlstr = (char *)urlString.String();
	BString		statustext;
	
/*	OLD WAY, DEPRECATED:
	// Set up Message to app manually
	BMessage	msg(B_ARGV_RECEIVED);
	msg.AddString("argv", "app");
	msg.AddString("argv", urlString);
	msg.AddInt32("argc", 2);
e.g.:	Launch("application/x-vnd.Be.URL.http", &msg, NULL);
*/
	// Launch url in preferred program
	if(type == "http") {
		roster_stat = be_roster->Launch("application/x-vnd.Be.URL.http", 1, &urlstr, NULL);
	} else if(type == "https") {
		roster_stat = be_roster->Launch("application/x-vnd.Be.URL.https", 1, &urlstr, NULL);
	} else if(type == "ftp") {
		roster_stat = be_roster->Launch("application/x-vnd.Be.URL.ftp", 1, &urlstr, NULL);
	} else if(type == "email") {
		roster_stat = be_roster->Launch("text/x-email", 1, &urlstr, NULL);
	} else if(type == "beshare") {
		// I'd like to have used:
		// application/x-vnd.Be.URL.beshare
		// ... but it doesn't seem to exist on some BeOS systems ??!
		roster_stat = be_roster->Launch("application/x-vnd.Sugoi-BeShare", 1, &urlstr, NULL);
	} else if(type == "rtsp") {
		roster_stat = be_roster->Launch("application/x-vnd.Be.URL.rtsp", 1, &urlstr, NULL);
	} else if(type == "news") {
		roster_stat = be_roster->Launch("application/x-vnd.Be.URL.news", 1, &urlstr, NULL);
	} else if(type == "telnet") {
		roster_stat = be_roster->Launch("application/URL: telnet", 1, &urlstr, NULL);
	} else if(type == "audiostream") {
		BMessage msg(B_REFS_RECEIVED);
		msg.AddString("be:url", urlString);
		roster_stat = be_roster->Launch("audio/x-scpls", &msg);
	} else { // failsafe decision is http
		roster_stat = be_roster->Launch("application/x-vnd.Be.URL.http", 1, &urlstr, NULL);
	}
	
	// Display Status
	// NetPositive is SingleLaunch, so B_ALREADY_RUNNING is returned == No Error!
	if(roster_stat == B_OK || roster_stat == B_ALREADY_RUNNING)
	{
		statustext << "Launching URL: " << type;
		DisplayStatus(statustext.String(), greenCol);
	} else {
		statustext << "Error launching " << type << " URL: ";
		
		if(roster_stat == B_BAD_VALUE) {
			statustext.Append("B_BAD_VALUE (the preferred app only accepts command-line arguments, no messages. Try a shell command instead.)");
		} else if(roster_stat == B_LAUNCH_FAILED) {
			statustext.Append("B_LAUNCH_FAILED (the preferred app failed to launch)");
		} else {
			statustext.Append("UNKNOWN ERROR. There could be a problem with the BeOS filetypes database.");
		}

		DisplayStatus(statustext.String(), redCol);
	}
	
	return(roster_stat);		
}



status_t ShellWindow::SearchTrackerFolder(BString pathString)
{
	// If path is encapsulated in quotes, remove them for now.
   	if(pathString.FindFirst("\"") != B_ERROR) pathString.RemoveAll("\"");

	if(pathString.FindFirst("dir://", 0) == 0) {
	/* --- is Trigger in first char ---*/
		pathString.RemoveFirst("dir://");
		
		if(pathString.CountChars() == 0) {
	/* --- is *only* the Trigger == just open current workDir after checking --- */
			if(WorkDirCheckSet(workDir, false) != B_OK) {
				DisplayStatus("Current workdir is invalid, resetting to homedir.", redCol);
				WorkDirCheckSet(homeDir, true);
				return B_OK;
			}
		} else {
	/* --- is a user-entered dir, check if valid --- */
			if(WorkDirCheckSet(pathString) != B_OK)
			{	// Trigger given but no valid Dir- print error message and return
				DisplayStatus("Error: Path is not a directory or doesn't exist.", redCol);
				return B_OK;
			}
		}
		// workDir now holds the correct path (see WorkDirCheckSet())
		OpenTrackerFolder(workDir);
		return B_OK;
	} else if(WorkDirCheckSet(pathString) == B_OK) {
	/* --- is plain command without a Trigger, if workdir B_OK open it --- */
		OpenTrackerFolder(workDir);
		return B_OK;
	}
	
	return B_ERROR;			// return (no open dir command found)
}


status_t ShellWindow::OpenTrackerFolder(BString pathString)
{
	BPath	trakpath;
	
	if(find_directory(B_BEOS_SYSTEM_DIRECTORY, &trakpath) == B_OK)
	{	// Build path to Tracker
		BString	trakcommand;
		trakpath.Append("Tracker ");
		// Encapsulate the path in quotes for Tracker
		pathString.Prepend("\"");
		pathString.Append("\"");
		trakcommand << trakpath.Path() << pathString.String() << " &";

		system( trakcommand.String() );
		
		DisplayStatus("Tracker folder opened, workdir changed.", greenCol);
		return B_OK;
	}
	return B_ERROR;
}




status_t ShellWindow::SearchFile(BString fileString)
{
	
	if(fileString.FindFirst("\"") != B_ERROR) fileString.RemoveAll("\""); // Strip off all quotes if any, or SetTo() will fail

	if(fileString.FindFirst("/", 0) != 0)
	{	// no "/" at beginning == relative from current workDir
		fileString.Prepend("/");
		fileString.Prepend(workDir);
	}

	BEntry		fileEntry;
	fileEntry.SetTo(fileString.String(), true);

	if(fileEntry.IsFile() == true)
	{	/* --- Check permissions --- */
		mode_t	perms;
		fileEntry.GetPermissions(&perms);
		
		if( (perms & S_IXUSR) || (perms & S_IXGRP) || (perms & S_IXOTH) )
		{	// File is executable, launch as shell command instead
			return B_ERROR;	// return(is NO file)
		} else {
		/* --- Launch file with Roster --- */
			entry_ref ref;
			if((fileEntry.GetRef(&ref)) == B_OK) {
				be_roster->AddToRecentDocuments(&ref, NULL);	// add to recent documents
				status_t	lstat = be_roster->Launch(&ref);
				if( (lstat != B_OK) && (lstat != B_ALREADY_RUNNING) ) {
					// If an error above, try to launch again with Tracker if active.
					if(be_roster->IsRunning(TRACKER_SIG) == true) {
						BPath	trakpath;
						if(find_directory(B_BEOS_SYSTEM_DIRECTORY, &trakpath) == B_OK) {
							trakpath.Append("Tracker ");
							BString	trakcommand;
							trakcommand << trakpath.Path() << "\"" << fileString.String() << "\" &";	
							system( trakcommand.String() );	
						}
					}
				}
			}
			DisplayStatus("Launching file with preferred app.", greenCol);
			return B_OK;	// return(is file)
		}
	}
	return B_ERROR;
}


int32 ShellWindow::ExecuteShellCommand(BString commandString)
{
	int32	shellExit = -1;
	bool	amperExist = false;

	/* --- Search for an amper ( &) in last 2 chars and remove --- */
	if( commandString.FindFirst(" &", (commandString.CountChars() - 2)) != B_ERROR ) {
		amperExist = true;
		commandString.RemoveLast(" &");
	}
	if(showStdout)
		commandString << " 2>&1";	// Merge stderr with stdout!
	if(addAmper)
		amperExist = true;
	if(amperExist)
		commandString << " &";	// (Re-)Append "&"
	
	/* --- Prepend a "cd" command (with quotes) to the workDir if it is valid --- */
	if(WorkDirCheckSet(workDir, false) != B_OK) {
		DisplayStatus("Current workdir is invalid: cancelling command and resetting to homedir.", redCol);
		WorkDirCheckSet(homeDir, true);
		return false;
	}
	BString	prepDir = "cd ";
	prepDir << "\"" << workDir << "\"" << "; ";
	commandString.Prepend(prepDir);

	#ifdef DEBUG
		cout << commandString.String() << endl;
	#endif
	
	if(showStdout) {
		BString	winTitle;
		if(showOutTitle) {
			winTitle = commandString.String();
			winTitle.RemoveAll(" 2>&1");
			winTitle.RemoveFirst(prepDir);	// Remove the "cd" command
		} else {
			winTitle = "stdout";
		}

		textWin = new TextWin(BRect(10,10,10,10), winTitle.String(), this);
		
		BMessage *startMsg = new BMessage(SHOWPROCESS);
		startMsg->AddString("command", commandString.String());
		textWin->PostMessage(startMsg);
		delete startMsg;
	} else {
		/* --- Spawn a new thread for the system function, and execute with sh --- */
		thread_id	shellThreadID = -1;		// Set to invalid value for the check below
		shellThreadID = spawn_thread((thread_func )ShellThread, "LaunchShellCommand", B_NORMAL_PRIORITY, (void *)commandString.String());
		if(shellThreadID >= 0)
			resume_thread(shellThreadID);

		/* --- Evaluate exit status, if possible --- */	
		if(amperExist == true) {	// No error checking possible!
			DisplayStatus("Executed shell command.", yelCol);
		} else {	// Wait until thread returns with status and do Error checking. 		// TODO: dies als gemeinsame funktion fürs text und shell win!!
			wait_for_thread(shellThreadID, &shellExit);
		
			if(shellExit == 127) {
				DisplayStatus("Error: program not found, exec failed.", redCol);
			} else if(shellExit == -1) {
				DisplayStatus("Error: got -1, couldn't create program process, fork failed.", redCol);
			} else if(WIFEXITED(shellExit) != 0) {	// prog exited normally with exit
				int progExit = WEXITSTATUS(shellExit);
				BString	stattext = "Shell command executed, returned ";
				stattext << (int)progExit;
		
				statusView->NewText(stattext.String());
				// Determine a Color for StatusBar and set
				if(progExit == EXIT_SUCCESS) {	// prog returned 0 (OK)
					statusView->NewCol(greenCol);
				} else {						// prog returned !=0 (error likely)
					statusView->NewCol(redCol);
				}
			} else if(WIFSIGNALED(shellExit) != 0) {
				DisplayStatus("Serious Error: program terminated because it received a signal that was not handled.", redCol);
			} else if(WIFSTOPPED(shellExit) != 0) {
				DisplayStatus("Serious Error: program was stopped.", redCol);
			} else if(WIFCORED(shellExit) != 0) {
				DisplayStatus("Serious Error: program terminated and produced a core dump.", redCol);
			}
		}

	}

	return(shellExit);
}


int32 ShellWindow::ShellThread(const char *command)
{
	return system(command);
}


void ShellWindow::OpenPrefWin( void )
{
	/* --- Check if a preferences window is already open --- */
	bool	prefWinOpen = false;
	int32	pwinNum;
	for(pwinNum=0; pwinNum<be_app->CountWindows(); pwinNum++) {
		if(is_kind_of(be_app->WindowAt(pwinNum), PrefWindow)) {
			prefWinOpen = true;
			break;	// found PrefWindow, quit the loop.
		}
	}
			
	/* --- If so, make it the focus window --- */
	if(prefWinOpen == true) {
		be_app->WindowAt(pwinNum)->Activate();
	} else {
	/* --- Make a new prefs window in the middle of the screen --- */
		BScreen *theScreen = new BScreen();
		BRect prefBox = theScreen->Frame();
		float theWidth = prefBox.right;
		float theHeight = prefBox.bottom;
		
		prefBox.left = (theWidth - PREF_WIDTH) / 2;
		prefBox.top = (theHeight - PREF_HEIGHT) / 2;
		prefBox.right = prefBox.left + PREF_WIDTH;
		prefBox.bottom = prefBox.top + PREF_HEIGHT;			
		
		PrefWindow	*thePrefWin;
		thePrefWin = new PrefWindow(prefBox, &prefs);
		thePrefWin->Show();
	}
}


/*======== Functions for updating the window after a preferences change ========*/

void ShellWindow::SetTheWinType(const char *winTypeChar)
{
	if( (strcmp(winTypeChar, "Standard")) == 0 )
		SetType(B_DOCUMENT_WINDOW);
	else if( (strcmp(winTypeChar, "Titled")) == 0 )
		SetType(B_TITLED_WINDOW);
	else if( (strcmp(winTypeChar, "Mini-Tab")) == 0 )
		SetLook(B_FLOATING_WINDOW_LOOK);
	else if( (strcmp(winTypeChar, "Tabless")) == 0 )
		SetLook(B_MODAL_WINDOW_LOOK);
	else if( (strcmp(winTypeChar, "Borderless")) == 0 )
		SetLook(B_BORDERED_WINDOW_LOOK);
	else	// for safety, i.e. when no setting file present
		SetType(DEF_WIN_TYPE);

	if(stayOnTop == true)	// Stay On Top needs to be set again now
		SetFeel(B_FLOATING_ALL_WINDOW_FEEL);
}


void ShellWindow::SetBubbleHelp(bool enable)
{
	showBubble = enable;
	bubbleMI->SetMarked(showBubble);
	
	// BeApp will create the BubbleHelper() for us if not existant but needed.
	if(showBubble == true) {	/* --- Show Bubble help --- */
		((LittleShellApp *)be_app)->SetBubbleHelp(menuBar, "Left-click and hold to move main window.");
		((LittleShellApp *)be_app)->SetBubbleHelp(fAView, "Drop files here with held-down modifier keys.");
		((LittleShellApp *)be_app)->SetBubbleHelp(okButt, "Click or press ENTER to execute command.");
		((LittleShellApp *)be_app)->SetBubbleHelp(histMenuField, "Command history menu - holds recent commands.\n"
			"Also accessible by Autocomplete and Up/Down Arrow keys.");
		((LittleShellApp *)be_app)->SetBubbleHelp(statusView, "Status messages are displayed here.\n\n"
			"Yellow - waiting for next action\n"
			"Green - successful operation\n"
			"Red - critical error occured");
		((LittleShellApp *)be_app)->SetBubbleHelp(commandBar, "Enter your command here:\n\n"
			"- Shell commands like in Terminal\n"
			"- URL's  [ mailto, ftp, http, https, www,\n"
			"  news, rtsp, telnet, audio, beshare ]\n"
			"- Directory path to open in a Tracker window\n"
			"- CD command to change work directory\n"
			"- Filename to open with preferred app");
	} else {					/* --- Empty the BubbleHelp tips --- */
		((LittleShellApp *)be_app)->SetBubbleHelp(menuBar, NULL);
		((LittleShellApp *)be_app)->SetBubbleHelp(fAView, NULL);
		((LittleShellApp *)be_app)->SetBubbleHelp(okButt, NULL);
		((LittleShellApp *)be_app)->SetBubbleHelp(histMenuField, NULL);
		((LittleShellApp *)be_app)->SetBubbleHelp(statusView, NULL);
		((LittleShellApp *)be_app)->SetBubbleHelp(commandBar, NULL);
	}
}


void ShellWindow::UpdateHistory(const char *commandString)
{
	hist_busy = true;
	
	// If the command is already on the list, remove it first
	for(int32 i=0; i<historyMenu->CountItems(); i++) {
		if( (strcmp(historyMenu->ItemAt(i)->Label(), commandString) == 0) ) {
			delete historyMenu->RemoveItem(i);
			commandBar->RemoveAutoItem(i);	// update history of OTextControl too!
		}
	}
	
	if(commandString) {
	if(commandString[0]) {
		// Add to autocomplete list of CommandBar as first item
		commandBar->AddAutoText(new BString(commandString), 0);
		// Remove ALT-L Shortcut from previous first MenuItem (if there still is an entry ;-))
		if(historyMenu->ItemAt(0))
			historyMenu->ItemAt(0)->SetShortcut((char )NULL, (uint32 )NULL);
		// Add the new command to the menu list as first menu item.
		BMessage *histmsg = new BMessage(HIST_ITEM_SELECTED);
		histmsg->AddString("history", commandString);
		historyMenu->AddItem(new BMenuItem(commandString, histmsg, 'L'), 0);
	}
	}
	
	// Restrict to maxHistItems number, delete oldest entry.
	int32	totalItems = historyMenu->CountItems();
	while(totalItems > maxHistItems) {
		// totalItems starts at 1, but list/menu index at 0, therefor "totalItems-1"
		delete historyMenu->RemoveItem(totalItems-1);
		commandBar->RemoveAutoItem(totalItems-1);
		totalItems--;
	}
	
	/*-----------------  Write history to settings file! ------------------*/
	
	prefs.SetString("lastcommand", commandString);
	if(saveWorkDir) prefs.SetString("lastworkdir", workDir.String());
	
	// Get the history from menu
	prefs.RemoveName("history");
	for(int32 i=0; i<historyMenu->CountItems(); i++) {
		prefs.AddString("history", historyMenu->ItemAt(i)->Label());
	}
	
	prefs.Save();
	
	// Set special bool to true (== history is not being updated)
	hist_busy = false;
}


void ShellWindow::UpdateMaxHist(int32 maxOptions)
{
	/*------------------------ Update menu -----------------------------*/
	
	// Update the maxHistItems private datamember
	maxHistItems = maxOptions;
	
	// Restrict to maxHistItems number, delete oldest entry.
	int32	totalItems = historyMenu->CountItems();
	while(totalItems > maxHistItems) {
		delete historyMenu->RemoveItem(totalItems-1);
		commandBar->RemoveAutoItem(totalItems-1);
		totalItems--;
	}

	/*------------------------- Save prefs -----------------------------*/
	prefs.RemoveName("history");

	// Get the history from menu and write to to settings file.
	for(int32 i=0; i<historyMenu->CountItems(); i++) {
		prefs.AddString("history", historyMenu->ItemAt(i)->Label());
	}

	prefs.SetInt32("maxhistory", maxHistItems);
	prefs.Save();
}


void ShellWindow::UpdateAddAmper( bool appendAmper )
{
	addAmper = appendAmper;
}


void ShellWindow::UpdateShowOut( bool showOut )
{
	showStdout = showOut;
	showOutMI->SetMarked(showStdout);
}


void ShellWindow::UpdateOutTitle(bool titleOut)
{
	showOutTitle = titleOut;
}


void ShellWindow::UpdateStayOnTop( bool stayTop )
{
	stayOnTop = stayTop;
	stayOnTopMI->SetMarked(stayOnTop);
	if(stayOnTop == true) {
		SetFeel(B_FLOATING_ALL_WINDOW_FEEL);
	} else {
		SetFeel(B_NORMAL_WINDOW_FEEL);
	}
}


void ShellWindow::UpdateAutoClose(bool winaClose)
{
	autoClose = winaClose;
	autoCloseMI->SetMarked(autoClose);
}


void ShellWindow::UpdateShowWDir(bool showIt)
{
	showWorkDir = showIt;
	
	BString newTitle = APP_NAME;
	newTitle << "  [" << (uint32)winNum << "]";
	if(showWorkDir == true) {
		newTitle << "  " << workDir.String();	
	}
	SetTitle(newTitle.String());
}


void ShellWindow::UpdateSaveWDir(bool saveIt)
{
	saveWorkDir = saveIt;
}
