/* ++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
/* +++               TextWindow.cpp                   +++ */
/* ++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
/* +++         Copyright © 2007 Ove Peter.            +++ */
/* ++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

#include <interface/TextView.h>
#include <interface/ScrollView.h>
#include <interface/Screen.h>
#include <app/Message.h>
#include <support/String.h>
#include <stdio.h>			// for popen()
#include <stdlib.h>			// for EXIT_SUCCESS, WIFEXITED, etc..
#include <fstream.h>		// for ifstream
#include <iostream>

#include "TextWindow.h"

TextWin::TextWin(BRect frame, const char *title, BHandler *target)
	: BWindow(frame, title, B_DOCUMENT_WINDOW, 0)
{
	// Init a pointer to the target BHandler (which likely called this window)
	caller = target;

	// Get the window Rect relative from parent Shell window , and resize/move
	BRect	winRect = ((BWindow *)caller)->Frame();	// BAD!
	BScreen theScreen;
	BRect	wholeScreen = theScreen.Frame();
	BRect	textWinRect(winRect.left, winRect.bottom+30.0, winRect.right, winRect.bottom+300.0);
	if(textWinRect.bottom > wholeScreen.Height())	// Bottom is out of the screen, move to top.
	{
		textWinRect.bottom = winRect.top-30.0;
		textWinRect.top = textWinRect.bottom-300.0;
	}
	ResizeTo(textWinRect.Width(), textWinRect.Height());
	MoveTo(textWinRect.left, textWinRect.top);
	
	// Add Text and Scroll Views
	BRect textframe = Bounds();
	textframe.right -= B_V_SCROLL_BAR_WIDTH;
	
	tView = new BTextView(textframe,
							"tview",
							textframe,
							B_FOLLOW_ALL_SIDES, B_WILL_DRAW);

	rgb_color	textBackCol = {240, 205, 0, 255};
	tView->SetFontAndColor(be_fixed_font, B_FONT_ALL, NULL);
	tView->SetViewColor(textBackCol);
	tView->MakeEditable(false);
	tView->SetWordWrap(true);
	
	BScrollView		*scrollView = new BScrollView("scroll_view",
									tView,
									B_FOLLOW_ALL_SIDES,
									0,
									false,
									true,
									B_FANCY_BORDER);
	AddChild(scrollView);
	tView->MakeFocus(true);

	// Do first Show() to setup message loop, but hide the window
	Minimize(true);
	Show();
}


TextWin::~TextWin()
{
	
}

/* ===================================================================== */

void TextWin::FrameResized(float width, float height)
{	// Wrap text
	BRect textrect = tView->TextRect();
	textrect.right = textrect.left + (width - B_V_SCROLL_BAR_WIDTH - 3.0);
	tView->SetTextRect(textrect);
}


void TextWin::MessageReceived(BMessage *message)
{
	switch(message->what)
	{	
	case SHOWFILE:
	{	// Show a textfile.
		const char	*filename;
		if( message->FindString("filename", &filename) == B_OK ) {
			DisplayFile(filename);
		}
	}
		break;

	case SHOWPROCESS:
	{	// Execute shell process and show stdout
		const char	*command;
		if( message->FindString("command", &command) == B_OK ) {
			DisplayShellProcess(command);
		}
	}
		break;

	default:
		BWindow::MessageReceived(message);
		break;
	}
}

/* ===================================================================== */

void TextWin::ReplaceText(const char *text)
{
	tView->SetText(text);
}


void TextWin::PrependText(const char *text)
{
	if(text) { // strlen() segfaults when feeding NULL
		tView->Insert(0, text, strlen(text));
		tView->ScrollToOffset(0);
	}
}


void TextWin::AppendText(const char *text)
{
	if(text) {
		tView->Insert(tView->TextLength(), text, strlen(text));
		tView->ScrollToOffset(tView->TextLength());
	}
}


void TextWin::CalcWindowHeight(void)
{	// If the text is lower than the window, make it smaller
	float text_height = tView->TextHeight( 0, tView->LineAt(tView->TextLength()) );
	if( text_height < Bounds().Height() ) {
		float win_height = (text_height + 5.0);
		float win_width = Bounds().Width();
		ResizeTo(win_width, win_height);
	}
}


status_t TextWin::DisplayFile(const char *filename)
{
	fstream	fin;	
	fin.open(filename, ios::in | ios::out);
	
	Minimize(false);
	
	if(fin.is_open() == 0) {	// error opening file, close win
		DisplayStatus("ERROR: cannot open file.", true);
		tView->SetText("ERROR: cannot open file\n");
		AppendText(filename);
		return B_ERROR;
	} else {					// Show whole file after it has been written
		BString	s;
		char	c;
		while(!fin.fail()) {	// WAS: !fin.eof() !!
			fin.get(c);
			s << c;
		}
		fin.close();			// Close filestream
		
		if(s.CountChars() > 1) {
			AppendText(s.String());
		} else {
			DisplayStatus("The file did not contain any text.", true);
			tView->SetText("The file did not contain any text.");
			return B_ERROR;
		}
		CalcWindowHeight();		// Recalculate window height		
		return B_OK;
	}
}


int TextWin::DisplayShellProcess(const char *command)
{
	int 		shellExit = -1;
	FILE		*fd;
	char		buff[256];

	fd = popen(command, "r");
	
	if(fd) {
		for(int32 i = 0; fgets(buff, 256, fd) != NULL; i++) {
			AppendText(buff);
			if(i <= 1) {	// only show win if any text in first 2 loops.
				if(buff != NULL) {
					Minimize(false);
				}
			}
		}	

		shellExit = pclose(fd);	// Close filepointer.
		CalcWindowHeight();		// Recalculate window height
		
		if(shellExit == 127) {
			DisplayStatus("Error: program not found, exec failed.", true);
			Quit();
		} else if(shellExit == -1) {
			DisplayStatus("Error: got -1, couldn't create program process, fork failed.", true);
			Quit();
		} else if(WIFEXITED(shellExit) != 0) {	// prog exited normally with exit
			int progExit = WEXITSTATUS(shellExit);
			BString	stattext = "Shell command executed, returned ";
			stattext << (int)progExit;

			if(progExit == EXIT_SUCCESS) {		// prog returned 0 (OK)
				DisplayStatus(stattext.String(), false);
			} else {							// prog returned !=0 (error likely)
				DisplayStatus(stattext.String(), true);
			}
		} else if(WIFSIGNALED(shellExit) != 0) {
			DisplayStatus("Serious Error: program terminated because it received a signal that was not handled.", true);
		} else if(WIFSTOPPED(shellExit) != 0) {
			DisplayStatus("Serious Error: program was stopped.", true);
		} else if(WIFCORED(shellExit) != 0) {
			DisplayStatus("Serious Error: program terminated and produced a core dump.", true);
		}
	} else {
		DisplayStatus("Internal error opening stdout stream!", true, false);
		Quit();	// Delete this window if nothing was shown.
	}
	
	// At last, check if any text was outputted at all
	if(tView->TextLength() == 0) {
		// Send stat msg to caller Win. Do here manually, as we don't want to change the stat col.
		BMessage *statMsg = new BMessage(STATUSTEXT);
		statMsg->AddString("stattext", ". No text on stdout.");
		statMsg->AddBool("append", true);
		BLooper *bloop = caller->Looper();
		(bloop) ? (BLooper *)bloop->PostMessage(statMsg, caller) : cerr << "No text on stdout.\n";
		delete statMsg;
		Lock();
		Quit();		// delete this win.
	}
	
	return shellExit;
}


void TextWin::DisplayStatus(const char *statText, bool is_error, bool append)
{	/* Prepare a BMessage which contains the status */
	BMessage *statusMsg = new BMessage(STATUSTEXT);

	statusMsg->AddString("stattext", statText);
	statusMsg->AddBool("iserror", is_error);
	statusMsg->AddBool("append", append);

	/* Get looper of target BHandler and post the message to it (if no looper present, post to stdout) */
	BLooper *bloop = caller->Looper();
	(bloop) ? (BLooper *)bloop->PostMessage(statusMsg, caller) : cerr << statText << "\n";

	delete statusMsg;
}
