/* ++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
/* +++               OTextControl.cpp                 +++ */
/* ++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
/* +++         Copyright © 2007 Ove Peter.            +++ */
/* ++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

#define BUILDING_OLIB

#include <support/String.h>
#include <interface/Window.h>	// for Window()

#include "OTextControl.h"
#include "Colors.h"

/* ---- TODO
	- add searching for matches inside history (???)
	- add the (currently unused) label parameter
	- add do_animation option (currently unused)
--- */

OTextControl::OTextControl(BRect frame, const char *name, const char *label, const char *text, BMessage *message, uint32 resizingMode, uint32 flags, bool do_autocomplete, bool do_animation)
	:BControl(frame, name, NULL, message, resizingMode, flags)
{
	modelMsg = message;
	
	is_enabled = true;
	
	/* --- Construct an OTextView object --- */
	BRect	tvRect = Bounds();
	tvRect.InsetBy(1.0, 1.0);	// Shrink rect
	tvRect.top += 2.0;
	tvRect.left += 2.0;
	BRect	textRect(2, 3, (Bounds().Width() - 2), (tvRect.Height() - 3));
	
	tView = new OTextView(tvRect,"tviewchild",
				textRect,
				B_FOLLOW_LEFT_RIGHT,
				B_WILL_DRAW | B_NAVIGABLE | B_FRAME_EVENTS | B_PULSE_NEEDED,
				this,
				do_autocomplete);
	
	/* --- Set TextControl-alike parameters for the Text View --- */

	// Next applies only for "growing" with the text, and is needed:
	tView->MakeResizable(true, NULL);
	// Adjust next parameters to your needs:
	tView->MakeEditable(true);
	tView->SetWordWrap(false);
	tView->MakeSelectable(true);
	tView->SetStylable(false);
	tView->SetDoesUndo(true); // only works when shortcut implemented in menu ?
	tView->SetAlignment(B_ALIGN_LEFT);

	/* --- Recalculate Height, as OTextView could have resized itself dependant on font --- */
	float	tViewHeight = tView->Bounds().Height();
	if(tViewHeight != tvRect.Height()) {	// Child resized itself, resize this view too
		ResizeTo(Bounds().Width(), (tViewHeight + 5));
	}
	
	/* --- Store the new dimension in datamembers needed by Drawing functions --- */
	vRect = Bounds();
	vwidth = vRect.Width();
	vheight = vRect.Height();

	AddChild(tView);
	
	/* --- If any text given in constructor, display it now. --- */
	if(text) {
		tView->SetText(text);
	}
}


OTextControl::~OTextControl( void )
{
	delete tView;
	
	/* --- Wait for the FocusDrawer thread to exit --- */
	keep_going = false;
	thread_info *tinfo = NULL;
	while(get_thread_info(focus_thread, tinfo) == B_OK)
		snooze(10000);
}

/* ============== Private functions for Focus Animation =============== */

int32 OTextControl::_focus(void *arg)
{
	((OTextControl *)arg)->focus_animation();
	return B_OK;
}

// No. of points the line moves in each loop:
#define INC 2
// Integers representing directions:
#define LT_RT 0
#define RT_RB 1
#define RB_LB 2
#define LB_LT 3
// Distance in Point from each corner to the end or start of the RoundRectnagle from Draw():
#define DIST 2

void OTextControl::focus_animation()
{
	// Set parameters for first start.
	keep_going = true;
	static int32	stage = LT_RT;
	BPoint			start = vRect.LeftTop();
	BPoint			end = start;
	int32 			blue = 0;
	
	/* --- Start thread loop --- */
	while(keep_going) {
		snooze(10000);
	
		if(is_enabled) {
			/* --- Looper must be locked. --- */
			if(!LockLooper())
				break;
			
			// This black line will always erase the last "moving" line.
			SetLowColor(blackCol);
			StrokeLine(start, end, B_SOLID_LOW);
			
			/* --- Calculate points for linedrawing to simulate an animation --- */
			if(tView->IsFocus()) {
				switch (stage) {
				case LT_RT:
					start.x += INC;
					end.x += (INC + 0.5); 	// end is "faster" than start on x.
					if(end.x >= (vwidth - DIST)) { // end arrived at corner
						end.x = (vwidth - DIST);	// (minus the space for the rounded rectangle in draw())
						// Wait for start to arrive too before changing direction.
						if(start.x >= (vwidth - DIST)) {
							start.x = vwidth;
							start.y += DIST;		// Change direction.
							end = start;
							stage = RT_RB;
						}
					}
					break;
				case RT_RB:
					start.y += INC;
					end.y += (INC / 2);		// end is "slower" than start on y.
					if(start.y >= (vheight - DIST)) { // start arrived at corner.
						start.y = (vheight - DIST);
						if(end.y >= (vheight - DIST)) {
							start.y = vheight;
							start.x -= DIST;
							end = start;
							stage = RB_LB;
						}
					}
					break;
				case RB_LB:
					start.x -= INC;
					end.x -= (INC + 0.5);
					// Currently, leftmost corner (vRect.left) is always 0. 
					// This will change when an optional label parameter
					// like in BTextControl is added to this class.
					if(end.x <= (vRect.left + DIST)) {
						end.x = (vRect.left + DIST);
						if(start.x <= (vRect.left + DIST)) {
							start.x = vRect.left;
							start.y -= DIST;
							end = start;
							stage = LB_LT;
						}
					}
					break;
				case LB_LT:
					start.y -= INC;
					end.y -= (INC / 2);
					// Topmost point is always 0, no calcs necessary :)
					if(start.y <= DIST) {
						start.y = DIST;
						if(end.y <= DIST) {
							start.y = 0;
							start.x += DIST;
							end = start;
							stage = LT_RT;
						}
					}
					break;
				}
				// Draw "moving" colored line.
				// SetLowColor(redCol);
				blue += 1;
				if(blue == 255) blue = 0;
				SetLowColor(255, 100, blue, 255);
				StrokeLine(start, end, B_SOLID_LOW);
			}	/* tView->IsFocus() */
							
			if(!keep_going) 
				break;
		
			Window()->Sync();
			UnlockLooper();
		}	/* is_enabled */
	}	/* keep_going */
}

/* =========================== Hook functions ======================= */

void OTextControl::AllAttached()
{
	SetViewColor(whiteCol);
	
	BControl::AllAttached();
	
	/* --- Spawn our Thread responsible for drawing a moving Dot --- */
	focus_thread = spawn_thread(_focus, "FocusAnimation", B_NORMAL_PRIORITY, (void *)this);
	resume_thread(focus_thread);
}


void OTextControl::WindowActivated(bool active)
{
	is_enabled = active;
	SetEnabled(is_enabled);	// If Window is deactivated, disable tView.

	// The BView version is empty in R5, but execute anyway for safety.
	BControl::WindowActivated(active);
}


void OTextControl::MakeFocus(bool flag)
{
	BControl::MakeFocus(flag);

	// Give TextView child the focus for typing.
	tView->MakeFocus(flag);
	
	if(flag) {	// If getting focus, select all text.
		tView->SelectAll();
	} else {	// If loosing focus, send model message.
		Invoke(modelMsg);
	}
}


void OTextControl::KeyDown(const char *bytes, int32 numBytes)
{	/* Pass everything that arrives here on to the OTextViewChild
	++ *without* running through BControl::KeyDown()
	++ This is here for safety only. Normally, when tView is focus, everything should go to it anyway.
	*/
	tView->KeyDown(bytes, numBytes);
}


void OTextControl::FrameResized(float width, float height)
{
	// Update dimensions datamembers here, used by Draw() and focus_animation().
	vRect = Bounds();
	vwidth = vRect.Width();
	vheight = vRect.Height();
	
	// Be Hook
	BControl::FrameResized(width, height);

	// Update view (note: could also use B_FULL_UPDATE_ON_RESIZE in view flags)
	Invalidate();
}


void OTextControl::Draw(BRect updateRect)
{
	float	orgPenSize = PenSize();	
	SetPenSize(1.0);
	BeginLineArray(13);
	
	// First darker Shade, X
	AddLine( BPoint(2.0, 1.0), BPoint((vwidth - 2.0), 1.0), greyShadow );
	// Darker Shade, Y
	AddLine( BPoint(1.0, 2.0), BPoint(1.0, (vheight - 2.0)), greyShadow );
	// Darker Points at each corner
	AddLine( BPoint(0.0, 1.0), BPoint(1.0, 0.0), greyShadow );					// upper left
	AddLine( BPoint(0.0, (vheight - 1.0)), BPoint(1.0, vheight), greyShadow );	// lower left
	AddLine( BPoint((vwidth - 1.0), 0.0), BPoint(vwidth, 1.0), greyShadow );	// upper right
	AddLine( BPoint((vwidth -1.0), vheight), BPoint(vwidth, (vheight - 1)), greyShadow );	// lower right
	// Lighter grey shade, X + Y
	AddLine( BPoint(2.0, 2.0), BPoint((vwidth - 1.0), 2.0), greyColShade );
	AddLine( BPoint(2.0, 3.0), BPoint(2.0, (vheight - 1.0)), greyColShade );
	// Standard Grey Points at each corner and one inside
	AddLine( BPoint(0.0, 0.0), BPoint(0.0, 0.0), greyCol );			// upper left
	AddLine( BPoint(0.0, vheight), BPoint(0.0, vheight), greyCol );	// lower left
	AddLine( BPoint(vwidth, 0.0), BPoint(vwidth, 0.0), greyCol );	// upper right
	AddLine( BPoint(vwidth, vheight), BPoint(vwidth, vheight), greyCol );	// lower right
	AddLine( BPoint(3.0, 3.0), BPoint(3.0, 3.0), greyCol );			// Shade inside upper left
	EndLineArray();
	
	// Main Black Rounded Shape
	SetHighColor(blackCol);
	StrokeRoundRect(vRect, 3.0, 3.0);
	// Restore Pen Size
	SetPenSize(orgPenSize);
	
	BControl::Draw(updateRect);
}

/* ============ Public functions for BTextControl compatibility ================== */

OTextView *OTextControl::TextView()
{
	return tView;
}


void OTextControl::SetEnabled(bool enabled)
{	
	is_enabled = enabled;
	
	tView->MakeEditable(is_enabled);

	if(is_enabled) {
		tView->SetViewColor(whiteCol);
		SetViewColor(whiteCol);
	} else {
		tView->SetViewColor(greyCol);
		SetViewColor(greyCol);
	}	
	Invalidate();
	tView->Invalidate();
}


status_t OTextControl::SetMessage(BMessage *message)
{
	modelMsg = message;
	
	return BControl::SetMessage(message);
}


void OTextControl::SetModificationMessage(BMessage *message)
{
	modelMsg = message;
	BControl::SetMessage(message);
}


BMessage *OTextControl::Message() const
{
	return modelMsg;
}


BMessage *OTextControl::ModificationMessage() const
{
	return modelMsg;
}


status_t OTextControl::Invoke(BMessage *message)
{
	modelMsg = message;
	
	return BControl::Invoke(message);
}

/* ============ Individual functions for autocomplete =============== */

void OTextControl::SetAutoComplete(bool doIt)
{
	tView->SetAutoComplete(doIt);
}


bool OTextControl::IsAutoComplete()
{
	return tView->IsAutoComplete();
}


void OTextControl::AddAutoText(BString *text, int32 index)
{
	tView->AddAutoText(text, index);
}


void OTextControl::AddAutoText(BString *text)
{
	tView->AddAutoText(text);
}


void OTextControl::RemoveAutoItem(int32 index)
{
	tView->RemoveAutoItem(index);
}

/* ============ Individual functions for manipulating Text =============== */

void OTextControl::AppendText(BString text)
{
	tView->AppendText(text);
	// Set cursor to end and scroll visible area to end
	int32 textLength = tView->TextLength();
	tView->Select(textLength, textLength);
	tView->ScrollToOffset(textLength);
}


void OTextControl::AppendText(const char *text)
{
	tView->AppendText(text);

	int32 textLength = tView->TextLength();
	tView->Select(textLength, textLength);
	tView->ScrollToOffset(textLength);
}


void OTextControl::PrependText(BString text)
{
	tView->PrependText(text);
	// Set cursor to start and scroll visible area to start
	tView->Select(0, 0);
	tView->ScrollToOffset(0);
}


void OTextControl::PrependText(const char *text)
{
	tView->PrependText(text);

	tView->Select(0, 0);
	tView->ScrollToOffset(0);
}


void OTextControl::SetText(BString text)
{
	tView->SetText(text.String());
}


void OTextControl::SetText(const char *text)
{
	tView->SetText(text);
}

const char *OTextControl::Text()
{
	return tView->Text();
}


void OTextControl::SelectAll()
{
	tView->SelectAll();
}

void OTextControl::SetAlignment(alignment forLabel, alignment forText)
{
	/* --- TODO:
		- add label
		- set alignment functionality for label
	--- */
		
	tView->SetAlignment(forText);
}
