/* ++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
/* +++                OTextView.cpp                   +++ */
/* ++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
/* +++         Copyright © 2007 Ove Peter.            +++ */
/* ++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

#define BUILDING_OLIB

#include <support/String.h>
#include <interface/Control.h>

#include "OTextView.h"


OTextView::OTextView(BRect frame, const char *name, BRect textRect, uint32 resizingMode, uint32 flags, BControl *container, bool do_autocomplete)
	:BTextView(frame, name, textRect, resizingMode, flags)
{
	// Setup defaults / arguments
	doAuto = do_autocomplete;
	if(container) {
		parent = container;
	}
	triggerNum = 2;
	currAutoItem = 0;	
	autoList = new BList(0);

	// Set Font to a reasonable size
	BFont	font(be_fixed_font);
	if(font.Size() > 16.0) {
		font.SetSize(16.0);
	} else if(font.Size() < 8.0) {
		font.SetSize(8.0);
	}
	SetFontAndColor(&font, B_FONT_ALL);

	// Recalculate Height dependant on font
	font_height fheight;
	font.GetHeight(&fheight);
	float theheight = (fheight.ascent + fheight.descent + fheight.leading);
	ResizeTo(frame.Width(), (theheight + 5.0));
}


OTextView::~OTextView( void )
{
	void *anItem;
	for ( int32 i = 0; (anItem = autoList->ItemAt(i)); i++ )
		delete anItem;

	delete autoList;
}

/* ===================== Hook functions ================================= */

void OTextView::KeyDown(const char *bytes, int32 numBytes)
{
	switch( *bytes )
	{	
		case B_TAB:
			/*
			++ Mimik behaviour of a BTextControl to switch focus and send model msg.
			++ Also prevents a literally TAB insert in the TextView.
			*/
			if(parent) { // un-focus the parent OTextControl (which will also send model msg)
				parent->MakeFocus(false);
			} else {
				MakeFocus(false);
			}
			break;
		
		case B_ENTER:
			/* 
			++ Mimik behaviour of a BTextControl to select all text
			++ and send model message when ENTER is pressed. 
			++ Update autoList on ENTER too.
			++ Also prevents a single newline '\n' in TextView.
			++ Note for LittleShell app: this never gets executed, as the default button catches ENTER !
			*/
			SelectAll();
			
			// invoke the invoke function :)) so that model message gets send.
			if(parent) parent->Invoke(parent->Message());
			
			if(doAuto) {
				// If command is empty, do not add it to autoList
				BString *thetext = new BString(Text());
				if(thetext->CountChars() == 0) break;				
				
				// If the command is already on the list, remove it first
				for(int32 i=0; i<autoList->CountItems(); i++) {
					BString *tmpItem = ((BString*)autoList->ItemAt(i));
					if(tmpItem->Compare(*thetext) == 0) RemoveAutoItem(i);
				}
				
				// Add the text to autocomplete list at first position
				autoList->AddItem((void*)thetext, 0);
			}
			break;
		
		/* --- This will mimik Up-Down-Arrow history like in Terminal --- */
		case B_UP_ARROW:
			if(doAuto) {
				currAutoItem++;
				if(currAutoItem > (autoList->CountItems() -1))
					currAutoItem = 0;
				// Set text to next item in history
				BString *tmpItem = ((BString*)autoList->ItemAt(currAutoItem));
				if( tmpItem != NULL )
					SetText(tmpItem->String());
				// Set cursor to end of text and scroll to end (for text longer than visible area)
				Select(TextLength(), TextLength());
				ScrollToOffset(TextLength());
			}
			break;

		case B_DOWN_ARROW:
			if(doAuto) {
				currAutoItem--;
				if(currAutoItem < 0)
					currAutoItem = (autoList->CountItems() - 1);
				
				BString *tmpItem = ((BString*)autoList->ItemAt(currAutoItem));
				if( tmpItem != NULL )
					SetText(tmpItem->String());
				
				Select(TextLength(), TextLength());
				ScrollToOffset(TextLength());
			}
			break;
		
		default:
			/* Reset the up/down "loop" */
			currAutoItem = 0;
			/* Execute BTextView's Hooks and eventually highlight */
			hl = false;	// hl will eventually be set to true by InsertText()

			BTextView::KeyDown(bytes, numBytes);

			if(hl)
				Select(hl_offset, hl_end);
			break;
	}
}


void OTextView::InsertText(const char *text, int32 length, int32 offset, const text_run_array *runs)
{
	BString replace = text;

	/*
	++ This will only autocomplete if
	++ a) auto is on
	++ b) only one char goes through here
	++ c) The minimum number of letters before autocomp. is reached
	++ d) Current offset is at end of control string.
	*/
	if( (doAuto == true) && (length == 1) && (TextLength() >= triggerNum) && (offset == TextLength()) ) {		
		BString currText = Text();
		currText << text;												// Append the new char
		BString *anItem;
		for( int32 i = 0; (anItem = (BString*)autoList->ItemAt(i)); i++ ) {
			if((anItem->Compare(currText, offset + length)) == 0) {		// Match found
				currAutoItem = i;	// for updown action
				
				hl_offset = (offset + 1);
				hl = true;
				BString theMatch(anItem->String());
				theMatch.CopyInto(replace, offset, anItem->Length());	// Copy only *new* text (without first letters) to final string.
				length = replace.Length();
				hl_end = anItem->Length();
				break;
			}
		}
	}
	
	/*
	++ If one or more characters are inserted, user possibly pasted clipboard text.
	++ Make sure that newlines '\n' get stripped from it.
	++ Pressing ENTER generating a newline is already prevented in KeyDown().
	*/
	if( (length >= 1) && (replace.FindFirst("\n") != B_ERROR) ) {
		replace.RemoveAll("\n");
	}
	
	/* 
	++ Call BTextView's InsertText() finally with the (possibly) modified text.
	*/
	BTextView::InsertText(replace.String(), length, offset, runs);
}


void OTextView::MakeFocus(bool focused)
{
	BTextView::MakeFocus(focused);
	
	if(!focused) {	// Let parent BControl send model msg
		if(parent) {
			parent->Invoke(parent->Message());
		}
	}
}

/* ============ Individual functions for autocomplete =============== */

void OTextView::SetAutoComplete(bool doIt)
{
	doAuto = doIt;
}


bool OTextView::IsAutoComplete()
{
	return doAuto;
}


void OTextView::AddAutoText(BString *text, int32 index)
{
	if(text)
		autoList->AddItem((void*)text, index);
}


void OTextView::AddAutoText(BString *text)
{
	if(text)
		autoList->AddItem((void*)text);
}


void OTextView::RemoveAutoItem(int32 index)
{	// TODO: is this data-delete safe???
	delete autoList->RemoveItem(index);
	/*
	delete autoList->ItemAt(index);
	autoList->RemoveItem(index);
	*/
}

/* ============ Individual functions for manipulating Text =============== */

void OTextView::AppendText(BString text)
{
	Insert(TextLength(), text.String(), text.Length(), NULL);
}


void OTextView::AppendText(const char *text)
{	// strlen() segfaults when feeding NULL, test pointer.
	if(text) Insert(TextLength(), text, strlen(text), NULL);
}


void OTextView::PrependText(BString text)
{
	Insert(0, text.String(), text.Length(), NULL);
}


void OTextView::PrependText(const char *text)
{
	if(text) Insert(0, text, strlen(text), NULL);
}
